from typing import Dict, List, Tuple, Union, Any, cast
import distutils.core  # pylint: disable = no-name-in-module, import-error
import os
import pkg_resources
import requirements

import cq.checker
import cq.utils

try:
	# To allow working `cq` with multiple versions of setuptools we need
	# to try import from extern package (which works for versions
	# > 20.0.0 but not >34.0.0,<36.0.0, if it is not working we import
	# another Version source code that worked in <39.0.0

	import pkg_resources.extern.packaging.version

	Version = Union[
		pkg_resources.extern.packaging.version.Version,
		pkg_resources.extern.packaging.version.LegacyVersion,
	]
	BASE_VERSION = pkg_resources.extern.packaging.version.Version
except ImportError:
	Version = Union[  # type: ignore
		pkg_resources.SetuptoolsVersion,  # type: ignore
		pkg_resources.SetuptoolsLegacyVersion,  # type: ignore
	]
	BASE_VERSION = pkg_resources._SetuptoolsVersionMixin  # type: ignore # pylint: disable = protected-access


class RequirementsvalidatorChecker(cq.checker.Checker):
	NAME = 'requirements-validator'
	DESCRIPTION = 'Validation of requirements.txt file (see https://pip.pypa.io/en/latest/reference/pip_check/)'

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		output: List[cq.checker.ResultLine] = []
		message, return_code = cq.utils.run_external_checker('pip', ['check'])

		if return_code:
			output.append(
				cq.checker.ResultLine(file = 'requirements.txt', line = None, message = message, is_error = True)
			)

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = None,
			return_code = return_code,
			output = output,
		)


class MinVersion(BASE_VERSION):  # type: ignore
	def __init__(self) -> None:
		self._key = self

	def __str__(self) -> str:
		return 'MIN'

	def __repr__(self) -> str:
		return "<Version('MIN')>"

	def __lt__(self, other: Any) -> bool:
		return True

	def __le__(self, other: Any) -> bool:
		return True

	def __gt__(self, other: Any) -> bool:
		return False

	def __ge__(self, other: Any) -> bool:
		return False

	def __eq__(self, other: Any) -> bool:
		return isinstance(other, self.__class__)

	def __ne__(self, other: Any) -> bool:
		return not self.__eq__(other)


class MaxVersion(BASE_VERSION):  # type: ignore
	def __init__(self) -> None:
		self._key = self

	def __str__(self) -> str:
		return 'MAX'

	def __repr__(self) -> str:
		return "<Version('MAX')>"

	def __lt__(self, other: Any) -> bool:
		return False

	def __le__(self, other: Any) -> bool:
		return False

	def __gt__(self, other: Any) -> bool:
		return True

	def __ge__(self, other: Any) -> bool:
		return True

	def __eq__(self, other: Any) -> bool:
		return isinstance(other, self.__class__)

	def __ne__(self, other: Any) -> bool:
		return not self.__eq__(other)


MIN_VERSION = MinVersion()
MAX_VERSION = MaxVersion()


class VersionRange:
	def __init__(
		self,
		lower: Version,
		upper: Version,
		lower_inclusive: bool = True,
		upper_inclusive: bool = False,
	) -> None:
		self.lower = lower
		self.upper = upper
		self.lower_inclusive = lower_inclusive
		self.upper_inclusive = upper_inclusive

	@classmethod
	def from_specs(cls, specs: List[Tuple[str, str]]) -> 'VersionRange':
		lower = MIN_VERSION  # type: Version
		upper = MAX_VERSION  # type: Version
		lower_inclusive = True
		upper_inclusive = True
		for sign, version_str in specs:
			if sign == '==':
				lower = pkg_resources.parse_version(version_str)
				upper = pkg_resources.parse_version(version_str)
			elif sign.startswith('>'):
				lower = pkg_resources.parse_version(version_str)
				lower_inclusive = sign.endswith('=')
			elif sign.startswith('<'):
				upper = pkg_resources.parse_version(version_str)
				upper_inclusive = sign.endswith('=')
		return cls(lower, upper, lower_inclusive, upper_inclusive)

	def __str__(self) -> str:
		if self.lower == self.upper:
			return f'=={self.lower}'
		first_sign = '>=' if self.lower_inclusive else '>'
		second_sign = '<=' if self.upper_inclusive else '<'
		return f'{first_sign}{self.lower},{second_sign}{self.upper}'

	def issubset(self, other: 'VersionRange') -> bool:
		if not isinstance(other, VersionRange):
			return NotImplemented

		if other.lower_inclusive or not self.lower_inclusive:
			lower_condition = other.lower <= self.lower
		else:
			lower_condition = other.lower < self.lower
		if other.upper_inclusive or not self.upper_inclusive:
			upper_condition = other.upper >= self.upper
		else:
			upper_condition = other.upper > self.upper

		# check that we overlap
		overlap_condition = other.upper >= self.lower and other.lower <= self.upper
		# mypy considers the condition as Any instead of bool
		return cast(bool, lower_condition and upper_condition and overlap_condition)


class RequirementsSetupCompatiblityChecker(cq.checker.Checker):
	NAME = 'requirements_setup_compatibility'
	DESCRIPTION = 'Validation of version compatibility in files setup.py and requirements.txt'

	SETUP_FILE = './setup.py'
	REQUIREMENTS_FILE = './requirements.txt'

	@staticmethod
	def _create_requirements_dict(source: str) -> Dict[str, List[Tuple[str, str]]]:
		result = {}
		for req in requirements.parse(source):
			result[req.name] = req.specs
		return result

	def _check_requirements_setup_compatibility(
		self,
		requirements_contents: str,
		setup_contents: str,
	) -> List[cq.checker.ResultLine]:
		'''
		Returns output messages for requirements that
		1) are present in requirements.txt and missing in setup.py
		2) their range in requirements.txt is not subset of range in setup.py
		'''
		output: List[cq.checker.ResultLine] = []
		requirements_dict = self._create_requirements_dict(requirements_contents)
		setup_dict = self._create_requirements_dict(setup_contents)

		for name, specs in requirements_dict.items():
			if name not in setup_dict:
				output.append(
					cq.checker.ResultLine(
						file = 'setup.py',
						line = None,
						message = f"setup.py: does not contain requirement '{name}' that is in requirements.txt",
						is_error = True,
					)
				)
			else:
				requirements_version_range = VersionRange.from_specs(specs)
				setup_version_range = VersionRange.from_specs(setup_dict[name])

				if not requirements_version_range.issubset(setup_version_range):
					output.append(
						cq.checker.ResultLine(
							file = 'setup.py',
							line = None,
							message = (
								f"setup.py: requirement '{name}' version in requirements.txt "
								f'({requirements_version_range}) does not lie in version '
								f'of setup.py ({setup_version_range})'
							),
							is_error = True,
						)
					)
		return output

	def _get_setup_py_requirements(self) -> str:
		distribution = distutils.core.run_setup(self.SETUP_FILE)
		# mypy doesn't know that distribution may contain install_requires
		setup_requirements = list(distribution.install_requires)  # type: ignore
		# get all requirements in extras_require
		# mypy complains that disribution does not have extras_require
		extras_require = distribution.extras_require if distribution.extras_require is not None else {}  # type: ignore
		setup_requirements += [
			requirement
			for extra_requirements in extras_require.values()
			for requirement in extra_requirements
		]
		return '\n'.join(setup_requirements)

	def _get_requirements_txt_requirements(self) -> str:
		with open(self.REQUIREMENTS_FILE) as f:
			return f.read()

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:  # pylint: disable = unused-argument
		# setup.py may miss for applications
		if not os.path.exists(self.SETUP_FILE):
			return cq.checker.CheckerResult(checker_name = self.NAME, help_line = None, return_code = 0, output = [])
		# warn if requirements file is missing
		if not os.path.exists(self.REQUIREMENTS_FILE):
			return cq.checker.CheckerResult(
				checker_name = self.NAME,
				help_line = None,
				return_code = 0,
				output = [
					cq.checker.ResultLine(
						file = self.REQUIREMENTS_FILE,
						line = None,
						message = f'Missing file {self.REQUIREMENTS_FILE}',
						is_error = False,
					)
				],
			)

		setup = self._get_setup_py_requirements()
		requirements = self._get_requirements_txt_requirements()

		output = self._check_requirements_setup_compatibility(requirements, setup)

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = None,
			return_code = 0 if not output else 1,
			output = output,
		)



from typing import Iterable, Iterator, List, Tuple, Any, IO, Union

import os
import pathlib
import subprocess


def run_external_checker(
	cmd: str,
	options: List[str],
	stderr: Union[None, int, IO[Any]] = subprocess.STDOUT,
) -> Tuple[str, int]:
	'''
	Run cmd in subprocess, sanitize stdout
	'''
	p = subprocess.Popen([cmd] + options, stdout = subprocess.PIPE, stderr = stderr)

	stdout: bytes
	stdout, _ = p.communicate()

	return stdout.decode('utf-8').strip(), p.returncode


def _is_any_path_part_hidden(path: str) -> bool:
	return any(part.startswith('.') and part not in {'.', '..'} for part in pathlib.Path(path).parts)


def get_path_generator(modules: Iterable[str], ignore_hidden: bool = True) -> Iterator[str]:
	'''
	Try to find any python file for pylint
	If path is already a python file or module continue
	Else, if directory, try to find any *.py file in the directory
	Continue recursively for all subdirectories
	'''
	for module in modules:
		if not (ignore_hidden and _is_any_path_part_hidden(module)):
			if module.endswith('.py'):
				yield module
			elif os.path.isdir(module):
				if os.path.exists(os.path.join(module, '__init__.py')):
					yield module
				else:
					yield from map(str, pathlib.Path(module).glob('*.py'))
				subdirs = [
					os.path.join(module, path)
					for path in os.listdir(module)
					if os.path.isdir(os.path.join(module, path))
				]
				if subdirs:
					yield from get_path_generator(subdirs)


def get_all_python_files(modules: Iterable[str], ignore_hidden: bool = True) -> Iterator[str]:
	for module in modules:
		if not (ignore_hidden and _is_any_path_part_hidden(module)):
			if os.path.isfile(module) and module.endswith('.py'):
				yield module
			elif os.path.isdir(module):
				for dirpath, _, filenames in os.walk(module):
					for file in filenames:
						path = os.path.join(dirpath, file)
						if not (ignore_hidden and _is_any_path_part_hidden(path)) and file.endswith('.py'):
							yield path



import os
from setuptools import setup

setup(
	name = 'ql-cq',
	author = 'Quantlane',
	author_email = 'code@quantlane.com',
	version = open('version.txt').read().strip(),
	url = 'https://gitlab.com/quantlane/meta/cq',
	license = 'Apache License, Version 2.0',
	long_description = open('README.md').read(),
	long_description_content_type = 'text/markdown',
	install_requires = [
		# all dependencies listed to lock the versions
		'pylint==2.3.1',
		'astroid==2.3.3',
		'isort==4.3.21',
		'mccabe==0.6.1',
		'typed-ast==1.4.0',
		'wrapt==1.11.2',
		'pyflakes-ext==2.0.0',
		'pyflakes==2.1.1',
		'mypy==0.770',
		'mypy-extensions==0.4.3',
		'typing-extensions>=3.7.4,<4.0.0',
		'bellybutton>=0.3.0,<0.4.0',
		'astpath[xpath]==0.6.1',
		'lxml==4.4.1',
		'click>=3.0,<8.0',
		'toolz>=0.8.2,<1.0.0',
		'requirements-parser>=0.1.0',
		'pip>=19.0.0',
	],
	packages = ['cq', 'cq.checkers'],
	package_data = {'cq': ['checkers/pylintrc', 'checkers/.bellybutton.yml']},
	entry_points = {'console_scripts': ['cq=cq.main:main']},
)



from typing import List
import os
import re

import cq.checker
import cq.utils


class BranchNameChecker(cq.checker.Checker):
	NAME = 'branch_name_check'
	DESCRIPTION = 'Checks, whether current branch name comply with Quantlane standards'

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		result = cq.checker.CheckerResult(checker_name = self.NAME, help_line = None, return_code = 0, output = [])

		try:
			branch_name = os.environ['CI_COMMIT_REF_NAME']
		except KeyError:
			branch_name, return_code = cq.utils.run_external_checker('git', ['symbolic-ref', 'HEAD', '--short'])

			if return_code != 0:
				return result

		if re.match('master|feature/.*|release/.*|fix/.*|version/.*', branch_name):
			return result

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = None,
			return_code = 1,
			output = [
				cq.checker.ResultLine(
					file = None,
					line = None,
					message = (
						f"Branch '{branch_name}' doesn't comply with Quantlane standarts. "
						f'It should be in form of: master, feature/.*, release/.*, fix/.*, version/.*'
					),
					is_error = True,
				)
			],
		)
